# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime
import json
import numbers
import os
import re
import threading

from awscli.compat import queue
from awscli.customizations.history.db import (
    DatabaseConnection,
    DatabaseHistoryHandler,
    DatabaseRecordReader,
    DatabaseRecordWriter,
    PayloadSerializer,
    RecordBuilder,
)
from awscli.testutils import FileCreator, mock, skip_if_windows, unittest
from tests import CaseInsensitiveDict


class FakeDatabaseConnection:
    def __init__(self):
        self.execute = mock.MagicMock()
        self.closed = False

    def close(self):
        self.closed = True


class TestGetHistoryDBFilename(unittest.TestCase):
    def setUp(self):
        self.files = FileCreator()

    def tearDown(self):
        self.files.remove_all()


class TestDatabaseConnection(unittest.TestCase):
    def setUp(self):
        self.files = FileCreator()

    def tearDown(self):
        self.files.remove_all()

    @mock.patch('awscli.compat.sqlite3.connect')
    def test_can_connect_to_argument_file(self, mock_connect):
        expected_location = os.path.expanduser(
            os.path.join('~', 'foo', 'bar', 'baz.db')
        )
        DatabaseConnection(expected_location)
        mock_connect.assert_called_with(
            expected_location, check_same_thread=False, isolation_level=None
        )

    @mock.patch('awscli.compat.sqlite3.connect')
    def test_does_try_to_enable_wal(self, mock_connect):
        conn = DatabaseConnection(':memory:')
        conn._connection.execute.assert_any_call('PRAGMA journal_mode=WAL')

    def test_does_ensure_table_created_first(self):
        db = DatabaseConnection(":memory:")
        cursor = db.execute('PRAGMA table_info(records)')
        schema = [col[:3] for col in cursor.fetchall()]
        expected_schema = [
            (0, 'id', 'TEXT'),
            (1, 'request_id', 'TEXT'),
            (2, 'source', 'TEXT'),
            (3, 'event_type', 'TEXT'),
            (4, 'timestamp', 'INTEGER'),
            (5, 'payload', 'TEXT'),
        ]
        self.assertEqual(expected_schema, schema)

    @mock.patch('awscli.compat.sqlite3.connect')
    def test_can_close(self, mock_connect):
        connection = mock.Mock()
        mock_connect.return_value = connection
        conn = DatabaseConnection(':memory:')
        conn.close()
        self.assertTrue(connection.close.called)

    @skip_if_windows('chmod is not supported on Windows')
    def test_sets_file_permissions_to_0600(self):
        db_path = os.path.join(self.files.rootdir, 'history.db')
        DatabaseConnection(db_path)
        permissions = os.stat(db_path).st_mode & 0o777
        self.assertEqual(permissions, 0o600)


class TestDatabaseHistoryHandler(unittest.TestCase):
    UUID_PATTERN = re.compile(
        '^[0-9a-f]{8}-([0-9a-f]{4}-){3}[0-9a-f]{12}$', re.I
    )

    def test_emit_does_write_cli_rc_record(self):
        writer = mock.Mock(DatabaseRecordWriter)
        record_builder = RecordBuilder()
        handler = DatabaseHistoryHandler(writer, record_builder)
        handler.emit('CLI_RC', 0, 'CLI')
        call = writer.write_record.call_args[0][0]
        self.assertEqual(
            call,
            {
                'command_id': mock.ANY,
                'event_type': 'CLI_RC',
                'payload': 0,
                'source': 'CLI',
                'timestamp': mock.ANY,
            },
        )
        self.assertTrue(self.UUID_PATTERN.match(call['command_id']))
        self.assertIsInstance(call['timestamp'], numbers.Number)

    def test_emit_does_write_cli_version_record(self):
        writer = mock.Mock(DatabaseRecordWriter)
        record_builder = RecordBuilder()
        handler = DatabaseHistoryHandler(writer, record_builder)
        handler.emit('CLI_VERSION', 'Version Info', 'CLI')
        call = writer.write_record.call_args[0][0]
        self.assertEqual(
            call,
            {
                'command_id': mock.ANY,
                'event_type': 'CLI_VERSION',
                'payload': 'Version Info',
                'source': 'CLI',
                'timestamp': mock.ANY,
            },
        )
        self.assertTrue(self.UUID_PATTERN.match(call['command_id']))
        self.assertIsInstance(call['timestamp'], numbers.Number)

    def test_emit_does_write_api_call_record(self):
        writer = mock.Mock(DatabaseRecordWriter)
        record_builder = RecordBuilder()
        handler = DatabaseHistoryHandler(writer, record_builder)
        payload = {'foo': 'bar'}
        handler.emit('API_CALL', payload, 'BOTOCORE')
        call = writer.write_record.call_args[0][0]
        self.assertEqual(
            call,
            {
                'command_id': mock.ANY,
                'request_id': mock.ANY,
                'event_type': 'API_CALL',
                'payload': payload,
                'source': 'BOTOCORE',
                'timestamp': mock.ANY,
            },
        )
        self.assertTrue(self.UUID_PATTERN.match(call['command_id']))
        self.assertTrue(self.UUID_PATTERN.match(call['request_id']))

    def test_emit_does_write_http_request_record(self):
        writer = mock.Mock(DatabaseRecordWriter)
        record_builder = RecordBuilder()
        handler = DatabaseHistoryHandler(writer, record_builder)
        payload = {'body': b'data'}
        # In order for an http_request to have a request_id it must have been
        # preceeded by an api_call record.
        handler.emit('API_CALL', '', 'BOTOCORE')
        handler.emit('HTTP_REQUEST', payload, 'BOTOCORE')
        call = writer.write_record.call_args[0][0]
        self.assertEqual(
            call,
            {
                'command_id': mock.ANY,
                'request_id': mock.ANY,
                'event_type': 'HTTP_REQUEST',
                'payload': payload,
                'source': 'BOTOCORE',
                'timestamp': mock.ANY,
            },
        )
        self.assertTrue(self.UUID_PATTERN.match(call['command_id']))
        self.assertTrue(self.UUID_PATTERN.match(call['request_id']))

    def test_emit_does_write_http_response_record(self):
        writer = mock.Mock(DatabaseRecordWriter)
        record_builder = RecordBuilder()
        handler = DatabaseHistoryHandler(writer, record_builder)
        payload = {'body': b'data'}
        # In order for an http_response to have a request_id it must have been
        # preceeded by an api_call record.
        handler.emit('API_CALL', '', 'BOTOCORE')
        handler.emit('HTTP_RESPONSE', payload, 'BOTOCORE')
        call = writer.write_record.call_args[0][0]
        self.assertEqual(
            call,
            {
                'command_id': mock.ANY,
                'request_id': mock.ANY,
                'event_type': 'HTTP_RESPONSE',
                'payload': payload,
                'source': 'BOTOCORE',
                'timestamp': mock.ANY,
            },
        )
        self.assertTrue(self.UUID_PATTERN.match(call['command_id']))
        self.assertTrue(self.UUID_PATTERN.match(call['request_id']))

    def test_emit_does_write_parsed_response_record(self):
        writer = mock.Mock(DatabaseRecordWriter)
        record_builder = RecordBuilder()
        handler = DatabaseHistoryHandler(writer, record_builder)
        payload = {'metadata': {'data': 'foobar'}}
        # In order for an http_response to have a request_id it must have been
        # preceeded by an api_call record.
        handler.emit('API_CALL', '', 'BOTOCORE')
        handler.emit('PARSED_RESPONSE', payload, 'BOTOCORE')
        call = writer.write_record.call_args[0][0]
        self.assertEqual(
            call,
            {
                'command_id': mock.ANY,
                'request_id': mock.ANY,
                'event_type': 'PARSED_RESPONSE',
                'payload': payload,
                'source': 'BOTOCORE',
                'timestamp': mock.ANY,
            },
        )
        self.assertTrue(self.UUID_PATTERN.match(call['command_id']))
        self.assertTrue(self.UUID_PATTERN.match(call['request_id']))


class BaseDatabaseRecordTester(unittest.TestCase):
    def assert_contains_lines_in_order(self, lines, contents):
        for line in lines:
            self.assertIn(line, contents)
            beginning = contents.find(line)
            contents = contents[(beginning + len(line)) :]


class BaseDatabaseRecordWriterTester(BaseDatabaseRecordTester):
    UUID_PATTERN = re.compile(
        '^[0-9a-f]{8}-([0-9a-f]{4}-){3}[0-9a-f]{12}$', re.I
    )

    def setUp(self):
        self.db = DatabaseConnection(':memory:')
        self.writer = DatabaseRecordWriter(self.db)


class TestDatabaseRecordWriter(BaseDatabaseRecordWriterTester):
    def setUp(self):
        super(TestDatabaseRecordWriter, self).setUp()

    def _read_last_record(self):
        cursor = self.db.execute('SELECT * FROM records')
        written_record = cursor.fetchone()
        return written_record

    def test_can_close(self):
        connection = mock.Mock()
        writer = DatabaseRecordWriter(connection)
        writer.close()
        self.assertTrue(connection.close.called)

    def test_can_write_record(self):
        self.writer.write_record(
            {
                'command_id': 'command',
                'event_type': 'FOO',
                'payload': 'bar',
                'source': 'TEST',
                'timestamp': 1234,
            }
        )

        # Now that we have verified the order of the fields in the insert
        # statement we can verify that the record values are in the correct
        # order in the tuple.
        # (command_id, request_id, source, event_type, timestamp, payload)
        written_record = self._read_last_record()
        self.assertEqual(
            written_record, ('command', None, 'TEST', 'FOO', 1234, '"bar"')
        )

    def test_commit_count_matches_write_count(self):
        records_to_write = 10
        for _ in range(records_to_write):
            self.writer.write_record(
                {
                    'command_id': 'command',
                    'event_type': 'foo',
                    'payload': '',
                    'source': 'TEST',
                    'timestamp': 1234,
                }
            )
        cursor = self.db.execute('SELECT COUNT(*) FROM records')
        record_count = cursor.fetchone()[0]

        self.assertEqual(record_count, records_to_write)

    def test_can_write_cli_version_record(self):
        self.writer.write_record(
            {
                'command_id': 'command',
                'event_type': 'CLI_VERSION',
                'payload': (
                    'aws-cli/1.11.184 Python/3.6.2 Darwin/15.6.0 '
                    'botocore/1.7.42'
                ),
                'source': 'TEST',
                'timestamp': 1234,
            }
        )
        written_record = self._read_last_record()

        self.assertEqual(
            written_record,
            (
                'command',
                None,
                'TEST',
                'CLI_VERSION',
                1234,
                '"aws-cli/1.11.184 Python/3.6.2 Darwin/15.6.0 botocore/1.7.42"',
            ),
        )

    def test_can_write_cli_arguments_record(self):
        self.writer.write_record(
            {
                'command_id': 'command',
                'event_type': 'CLI_ARGUMENTS',
                'payload': ['s3', 'ls'],
                'source': 'TEST',
                'timestamp': 1234,
            }
        )

        written_record = self._read_last_record()
        self.assertEqual(
            written_record,
            ('command', None, 'TEST', 'CLI_ARGUMENTS', 1234, '["s3", "ls"]'),
        )

    def test_can_write_api_call_record(self):
        self.writer.write_record(
            {
                'command_id': 'command',
                'event_type': 'API_CALL',
                'payload': {
                    'service': 's3',
                    'operation': 'ListBuckets',
                    'params': {},
                },
                'source': 'TEST',
                'timestamp': 1234,
            }
        )

        written_record = self._read_last_record()
        self.assertEqual(
            written_record,
            (
                'command',
                None,
                'TEST',
                'API_CALL',
                1234,
                json.dumps(
                    {
                        'service': 's3',
                        'operation': 'ListBuckets',
                        'params': {},
                    }
                ),
            ),
        )

    def test_can_write_http_request_record(self):
        self.writer.write_record(
            {
                'command_id': 'command',
                'event_type': 'HTTP_REQUEST',
                'payload': {
                    'method': 'GET',
                    'headers': CaseInsensitiveDict({}),
                    'body': '...',
                },
                'source': 'TEST',
                'timestamp': 1234,
            }
        )

        written_record = self._read_last_record()
        self.assertEqual(
            written_record,
            (
                'command',
                None,
                'TEST',
                'HTTP_REQUEST',
                1234,
                json.dumps(
                    {
                        'method': 'GET',
                        'headers': {},
                        'body': '...',
                    }
                ),
            ),
        )

    def test_can_write_http_response_record(self):
        self.writer.write_record(
            {
                'command_id': 'command',
                'event_type': 'HTTP_RESPONSE',
                'payload': {
                    'streaming': False,
                    'headers': {},
                    'body': '...',
                    'status_code': 200,
                    'request_id': '1234abcd',
                },
                'source': 'TEST',
                'timestamp': 1234,
            }
        )

        written_record = self._read_last_record()
        self.assertEqual(
            written_record,
            (
                'command',
                None,
                'TEST',
                'HTTP_RESPONSE',
                1234,
                json.dumps(
                    {
                        'streaming': False,
                        'headers': {},
                        'body': '...',
                        'status_code': 200,
                        'request_id': '1234abcd',
                    }
                ),
            ),
        )

    def test_can_write_parsed_response_record(self):
        self.writer.write_record(
            {
                'command_id': 'command',
                'event_type': 'PARSED_RESPONSE',
                'payload': {},
                'source': 'TEST',
                'timestamp': 1234,
            }
        )

        written_record = self._read_last_record()
        self.assertEqual(
            written_record,
            ('command', None, 'TEST', 'PARSED_RESPONSE', 1234, '{}'),
        )

    def test_can_write_cli_rc_record(self):
        self.writer.write_record(
            {
                'command_id': 'command',
                'event_type': 'CLI_RC',
                'payload': 0,
                'source': 'TEST',
                'timestamp': 1234,
            }
        )

        written_record = self._read_last_record()
        self.assertEqual(
            written_record, ('command', None, 'TEST', 'CLI_RC', 1234, '0')
        )


class ThreadedRecordBuilder:
    def __init__(self, tracker):
        self._read_q = queue.Queue()
        self._write_q = queue.Queue()
        self._thread = threading.Thread(
            target=self._threaded_request_tracker, args=(tracker,)
        )

    def _threaded_request_tracker(self, builder):
        while True:
            event_type = self._read_q.get()
            if event_type is False:
                return
            payload = {'body': b''}
            request_id = builder.build_record(event_type, payload, '')
            self._write_q.put_nowait(request_id)

    def read_n_results(self, n):
        records = [self._write_q.get() for _ in range(n)]
        return records

    def request_id_for_event(self, event_type):
        self._read_q.put_nowait(event_type)

    def start(self):
        self._thread.start()

    def close(self):
        self._read_q.put_nowait(False)
        self._thread.join()


class TestMultithreadRequestId(unittest.TestCase):
    def setUp(self):
        self.builder = RecordBuilder()
        self.threads = []

    def tearDown(self):
        for t in self.threads:
            t.close()

    def start_n_threads(self, n):
        for _ in range(n):
            t = ThreadedRecordBuilder(self.builder)
            t.start()
            self.threads.append(t)

    def test_each_thread_has_separate_request_id(self):
        self.start_n_threads(2)
        self.threads[0].request_id_for_event('API_CALL')
        self.threads[0].request_id_for_event('HTTP_REQUEST')
        self.threads[1].request_id_for_event('API_CALL')
        self.threads[1].request_id_for_event('HTTP_REQUEST')

        a_records = self.threads[0].read_n_results(2)
        b_records = self.threads[1].read_n_results(2)

        # Each thread should have its own set of request_ids so the request
        # ids in each set of records should match, but should not match the
        # request_ids from the other thread.
        a_request_ids = [record['request_id'] for record in a_records]
        self.assertEqual(len(a_request_ids), 2)
        self.assertEqual(a_request_ids[0], a_request_ids[1])
        thread_a_request_id = a_request_ids[0]

        b_request_ids = [record['request_id'] for record in b_records]
        self.assertEqual(len(b_request_ids), 2)
        self.assertEqual(b_request_ids[0], b_request_ids[1])
        thread_b_request_id = b_request_ids[0]

        # Since the request_id is reset by the API_CALL record being written
        # and thread b has now written an API_CALL record the request id has
        # been reset once. To ensure this doesnt bleed over to other threads
        # we will write another record in thread a and ensure that it's
        # request_id matches the previous ones from thread a rather than then
        # thread b request_id
        self.threads[0].request_id_for_event('HTTP_RESPONSE')
        self.threads[1].request_id_for_event('HTTP_RESPONSE')
        a_record = self.threads[0].read_n_results(1)[0]
        b_record = self.threads[1].read_n_results(1)[0]
        self.assertEqual(a_record['request_id'], thread_a_request_id)
        self.assertEqual(b_record['request_id'], thread_b_request_id)


class TestDatabaseRecordReader(BaseDatabaseRecordTester):
    def setUp(self):
        self.fake_connection = FakeDatabaseConnection()
        self.reader = DatabaseRecordReader(self.fake_connection)

    def test_can_close(self):
        self.reader.close()
        self.assertTrue(self.fake_connection.closed)

    def test_row_factory_set(self):
        self.assertEqual(
            self.fake_connection.row_factory, self.reader._row_factory
        )

    def test_iter_latest_records_performs_correct_query(self):
        expected_query = (
            '    SELECT * FROM records\n'
            '        WHERE id =\n'
            '        (SELECT id FROM records WHERE timestamp =\n'
            '        (SELECT max(timestamp) FROM records)) ORDER BY timestamp;'
        )
        [_ for _ in self.reader.iter_latest_records()]
        self.assertEqual(
            self.fake_connection.execute.call_args[0][0].strip(),
            expected_query.strip(),
        )

    def test_iter_latest_records_does_iter_records(self):
        records_to_get = [1, 2, 3]
        self.fake_connection.execute.return_value.__iter__.return_value = iter(
            records_to_get
        )
        records = [r for r in self.reader.iter_latest_records()]
        self.assertEqual(records, records_to_get)

    def test_iter_records_performs_correct_query(self):
        expected_query = (
            'SELECT * from records where id = ? ' 'ORDER BY timestamp'
        )
        [_ for _ in self.reader.iter_records('fake_id')]
        self.assertEqual(
            self.fake_connection.execute.call_args[0][0].strip(),
            expected_query.strip(),
        )

    def test_iter_records_does_iter_records(self):
        records_to_get = [1, 2, 3]
        self.fake_connection.execute.return_value.__iter__.return_value = iter(
            records_to_get
        )
        records = [r for r in self.reader.iter_records('fake_id')]
        self.assertEqual(records, records_to_get)


class TestPayloadSerialzier(unittest.TestCase):
    def test_can_serialize_basic_types(self):
        original = {
            'string': 'foo',
            'int': 4,
            'list': [1, 2, 'bar'],
            'dict': {'sun': 'moon'},
            'float': 1.2,
        }
        string_value = json.dumps(original, cls=PayloadSerializer)
        reloaded = json.loads(string_value)
        self.assertEqual(original, reloaded)

    def test_can_serialize_datetime(self):
        now = datetime.datetime.now()
        iso_now = now.isoformat()
        string_value = json.dumps(now, cls=PayloadSerializer)
        reloaded = json.loads(string_value)
        self.assertEqual(iso_now, reloaded)

    def test_can_serialize_case_insensitive_dict(self):
        original = CaseInsensitiveDict({'fOo': 'bar'})
        string_value = json.dumps(original, cls=PayloadSerializer)
        reloaded = json.loads(string_value)
        self.assertEqual(original, reloaded)

    def test_can_serialize_unknown_type(self):
        original = queue.Queue()
        encoded = repr(original)
        string_value = json.dumps(original, cls=PayloadSerializer)
        reloaded = json.loads(string_value)
        self.assertEqual(encoded, reloaded)

    def test_can_serialize_non_utf_8_bytes_type(self):
        original = b'\xfe\xed'  # Non utf-8 byte sequence
        encoded = '<Byte sequence>'
        string_value = json.dumps(original, cls=PayloadSerializer)
        reloaded = json.loads(string_value)
        self.assertEqual(encoded, reloaded)

    def test_does_preserve_utf_8_bytes_type(self):
        original = b'foobar'  # utf-8 byte sequence
        encoded = 'foobar'
        string_value = json.dumps(original, cls=PayloadSerializer)
        reloaded = json.loads(string_value)
        self.assertEqual(encoded, reloaded)

    def test_can_serialize_non_utf_8_bytes_type_in_dict(self):
        original = {'foo': 'bar', 'bytes': b'\xfe\xed'}
        encoded = {'foo': 'bar', 'bytes': '<Byte sequence>'}
        string_value = json.dumps(original, cls=PayloadSerializer)
        reloaded = json.loads(string_value)
        self.assertEqual(encoded, reloaded)

    def test_can_serialize_non_utf_8_bytes_type_in_list(self):
        original = ['foo', b'\xfe\xed']
        encoded = ['foo', '<Byte sequence>']
        string_value = json.dumps(original, cls=PayloadSerializer)
        reloaded = json.loads(string_value)
        self.assertEqual(encoded, reloaded)

    def test_can_serialize_non_utf_8_bytes_type_in_tuple(self):
        original = ('foo', b'\xfe\xed')
        encoded = ['foo', '<Byte sequence>']
        string_value = json.dumps(original, cls=PayloadSerializer)
        reloaded = json.loads(string_value)
        self.assertEqual(encoded, reloaded)

    def test_can_serialize_non_utf_8_bytes_nested(self):
        original = {
            'foo': 'bar',
            'bytes': b'\xfe\xed',
            'list': ['foo', b'\xfe\xed'],
            'more_nesting': {
                'bytes': b'\xfe\xed',
                'tuple': ('bar', 'baz', b'\xfe\xed'),
            },
        }
        encoded = {
            'foo': 'bar',
            'bytes': '<Byte sequence>',
            'list': ['foo', '<Byte sequence>'],
            'more_nesting': {
                'bytes': '<Byte sequence>',
                'tuple': ['bar', 'baz', '<Byte sequence>'],
            },
        }
        string_value = json.dumps(original, cls=PayloadSerializer)
        reloaded = json.loads(string_value)
        self.assertEqual(encoded, reloaded)


class TestRecordBuilder(unittest.TestCase):
    UUID_PATTERN = re.compile(
        '^[0-9a-f]{8}-([0-9a-f]{4}-){3}[0-9a-f]{12}$', re.I
    )

    def setUp(self):
        self.builder = RecordBuilder()

    def _get_request_id_for_event_type(self, event_type):
        record = self.builder.build_record(event_type, {'body': b''}, '')
        return record.get('request_id')

    def test_does_inject_timestamp(self):
        record = self.builder.build_record('TEST', '', '')
        self.assertTrue('timestamp' in record)
        self.assertTrue(isinstance(record['timestamp'], numbers.Number))

    def test_does_inject_command_id(self):
        record = self.builder.build_record('TEST', '', '')
        self.assertTrue('timestamp' in record)
        self.assertTrue(isinstance(record['timestamp'], numbers.Number))
        self.assertTrue('command_id' in record)
        self.assertTrue(self.UUID_PATTERN.match(record['command_id']))

    def test_does_create_record_with_correct_fields(self):
        record = self.builder.build_record('type', 'payload', 'source')
        self.assertEqual(record['event_type'], 'type')
        self.assertEqual(record['payload'], 'payload')
        self.assertEqual(record['source'], 'source')
        self.assertTrue('command_id' in record)
        self.assertTrue('timestamp' in record)

    def test_can_process_http_request_with_none_body(self):
        try:
            self.builder.build_record('HTTP_REQUEST', {'body': None}, '')
        except ValueError:
            self.fail("Should not raise value error")

    def test_can_process_http_response_with_nono_body(self):
        try:
            self.builder.build_record('HTTP_RESPONSE', {'body': None}, '')
        except ValueError:
            self.fail("Should not raise value error")

    def test_can_get_request_id_from_api_call(self):
        identifier = self._get_request_id_for_event_type('API_CALL')
        self.assertTrue(self.UUID_PATTERN.match(identifier))

    def test_does_get_id_for_http_request_with_api_call(self):
        call_identifier = self._get_request_id_for_event_type('API_CALL')
        request_identifier = self._get_request_id_for_event_type(
            'HTTP_REQUEST'
        )

        self.assertEqual(call_identifier, request_identifier)
        self.assertTrue(self.UUID_PATTERN.match(call_identifier))

    def test_does_get_id_for_http_response_with_api_call(self):
        call_identifier = self._get_request_id_for_event_type('API_CALL')
        response_identifier = self._get_request_id_for_event_type(
            'HTTP_RESPONSE'
        )

        self.assertEqual(call_identifier, response_identifier)
        self.assertTrue(self.UUID_PATTERN.match(call_identifier))

    def test_does_get_id_for_parsed_response_with_api_call(self):
        call_identifier = self._get_request_id_for_event_type('API_CALL')
        response_identifier = self._get_request_id_for_event_type(
            'PARSED_RESPONSE'
        )

        self.assertEqual(call_identifier, response_identifier)
        self.assertTrue(self.UUID_PATTERN.match(call_identifier))

    def test_does_not_get_id_for_http_request_without_api_call(self):
        identifier = self._get_request_id_for_event_type('HTTP_REQUEST')
        self.assertIsNone(identifier)

    def test_does_not_get_id_for_http_response_without_api_call(self):
        identifier = self._get_request_id_for_event_type('HTTP_RESPONSE')
        self.assertIsNone(identifier)

    def test_does_not_get_id_for_parsed_response_without_api_call(self):
        identifier = self._get_request_id_for_event_type('PARSED_RESPONSE')
        self.assertIsNone(identifier)


class TestIdentifierLifecycles(unittest.TestCase):
    def setUp(self):
        self.builder = RecordBuilder()

    def _get_multiple_request_ids(self, events):
        fake_payload = {'body': b''}
        request_ids = [
            self.builder.build_record(event, fake_payload.copy(), '')[
                'request_id'
            ]
            for event in events
        ]
        return request_ids

    def test_multiple_http_lifecycle_writes_have_same_request_id(self):
        request_ids = self._get_multiple_request_ids(
            ['API_CALL', 'HTTP_REQUEST', 'HTTP_RESPONSE', 'PARSED_RESPONSE']
        )
        # All request_ids should match since this is one request lifecycle
        unique_request_ids = set(request_ids)
        self.assertEqual(len(unique_request_ids), 1)

    def test_request_id_reset_on_api_call(self):
        request_ids = self._get_multiple_request_ids(
            [
                'API_CALL',
                'HTTP_REQUEST',
                'HTTP_RESPONSE',
                'PARSED_RESPONSE',
                'API_CALL',
                'HTTP_REQUEST',
                'HTTP_RESPONSE',
                'PARSED_RESPONSE',
                'API_CALL',
                'HTTP_REQUEST',
                'HTTP_RESPONSE',
                'PARSED_RESPONSE',
            ]
        )

        # There should be three distinct requet_ids since there are three
        # distinct calls that end with a parsed response.
        unique_request_ids = set(request_ids)
        self.assertEqual(len(unique_request_ids), 3)

        # Check that the request ids match the correct events.
        first_request_ids = request_ids[:4]
        unique_first_request_ids = set(first_request_ids)
        self.assertEqual(len(unique_first_request_ids), 1)

        second_request_ids = request_ids[4:8]
        unique_second_request_ids = set(second_request_ids)
        self.assertEqual(len(unique_second_request_ids), 1)

        third_request_ids = request_ids[8:]
        unique_third_request_ids = set(third_request_ids)
        self.assertEqual(len(unique_third_request_ids), 1)



# Copyright 2025 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from tests import HTTPResponse


class TestCreateRoleAssociationsCommand:
    def assert_call_matches(
        self, aws_request, service_account, cluster_name, namespace, role_arn
    ):
        assert aws_request.service_name == "eks"
        assert aws_request.operation_name == "CreatePodIdentityAssociation"
        assert aws_request.params["clusterName"] == cluster_name
        assert aws_request.params["namespace"] == namespace
        assert aws_request.params["roleArn"] == role_arn
        assert aws_request.params["serviceAccount"] == service_account

    def assert_roll_back(self, aws_request, cluster_name, associationId):
        assert aws_request.service_name == "eks"
        assert aws_request.operation_name == "DeletePodIdentityAssociation"
        assert aws_request.params["clusterName"] == cluster_name
        assert aws_request.params["associationId"] == associationId

    # Use case: Expect to return create pod identity association results for start job run
    # Expected results: Operation is performed by client
    # to create pod identity associations with start job run service accounts
    def test_create_role_associations_for_start_job_run(
        self,
        cli_runner,
        role_arn,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        create_pod_identity_association_response,
        start_job_run_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        result = cli_runner.run(
            [
                "emr-containers",
                "create-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
            ]
        )

        assert result.aws_requests[0].service_name == "eks"
        assert result.aws_requests[0].operation_name == "DescribeCluster"
        assert (
            len(result.aws_requests) == len(start_job_run_service_accounts) + 1
        )
        for i in range(len(start_job_run_service_accounts)):
            self.assert_call_matches(
                result.aws_requests[i + 1],
                start_job_run_service_accounts[i],
                cluster_name,
                namespace,
                role_arn,
            )

    # Use case: Expect to return create pod identity association results for interactive endpoint
    # Expected results: Operation is performed by client
    # to create pod identity associations with interactive endpoint service accounts
    def test_create_role_associations_for_interactive_endpoint(
        self,
        cli_runner,
        role_arn,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        create_pod_identity_association_response,
        interactive_endpoint_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        result = cli_runner.run(
            [
                "emr-containers",
                "create-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
                "--type",
                "interactive_endpoint",
            ]
        )

        assert result.aws_requests[0].service_name == "eks"
        assert result.aws_requests[0].operation_name == "DescribeCluster"
        assert (
            len(result.aws_requests)
            == len(interactive_endpoint_service_accounts) + 1
        )
        for i in range(len(interactive_endpoint_service_accounts)):
            self.assert_call_matches(
                result.aws_requests[i + 1],
                interactive_endpoint_service_accounts[i],
                cluster_name,
                namespace,
                role_arn,
            )

    # Use case: Expect to return create pod identity association results for spark operator
    # Expected results: Operation is performed by client
    # to create pod identity associations with spark operator service accounts
    def test_create_role_associations_for_spark_operator(
        self,
        cli_runner,
        role_arn,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        create_pod_identity_association_response,
        spark_operator_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        result = cli_runner.run(
            [
                "emr-containers",
                "create-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
                "--type",
                "spark_operator",
            ]
        )

        assert result.aws_requests[0].service_name == "eks"
        assert result.aws_requests[0].operation_name == "DescribeCluster"
        assert (
            len(result.aws_requests)
            == len(spark_operator_service_accounts) + 1
        )
        for i in range(len(spark_operator_service_accounts)):
            self.assert_call_matches(
                result.aws_requests[i + 1],
                spark_operator_service_accounts[i],
                cluster_name,
                namespace,
                role_arn,
            )

    # Use case: Expect to return create pod identity association results for spark operator with operator namespace
    # Expected results: Operation is performed by client
    # to create pod identity associations with spark operator service accounts in correct namespaces
    def test_create_role_associations_for_spark_operator_namespace(
        self,
        cli_runner,
        role_arn,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        create_pod_identity_association_response,
        spark_operator_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        result = cli_runner.run(
            [
                "emr-containers",
                "create-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
                "--type",
                "spark_operator",
                "--operator-namespace",
                "spark-operator",
            ]
        )
        assert result.aws_requests[0].service_name == "eks"
        assert result.aws_requests[0].operation_name == "DescribeCluster"
        assert (
            len(result.aws_requests)
            == len(spark_operator_service_accounts) + 1
        )
        for i in range(len(spark_operator_service_accounts)):
            ns = "spark-operator" if i == 0 else namespace
            self.assert_call_matches(
                result.aws_requests[i + 1],
                spark_operator_service_accounts[i],
                cluster_name,
                ns,
                role_arn,
            )

    # Use case: Expect to return create pod identity association results for flink operator
    # Expected results: Operation is performed by client
    # to create pod identity associations with flink operator service accounts
    def test_create_role_associations_for_flink_operator(
        self,
        cli_runner,
        role_arn,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        create_pod_identity_association_response,
        flink_operator_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        result = cli_runner.run(
            [
                "emr-containers",
                "create-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
                "--type",
                "flink_operator",
            ]
        )

        assert result.aws_requests[0].service_name == "eks"
        assert result.aws_requests[0].operation_name == "DescribeCluster"
        assert (
            len(result.aws_requests)
            == len(flink_operator_service_accounts) + 1
        )
        for i in range(len(flink_operator_service_accounts)):
            self.assert_call_matches(
                result.aws_requests[i + 1],
                flink_operator_service_accounts[i],
                cluster_name,
                namespace,
                role_arn,
            )

    # Use case: Expect to return create pod identity association results for flink operator with operator namespaces
    # Expected results: Operation is performed by client
    # to create pod identity associations with flink operator service accounts in correct namespace
    def test_create_role_associations_for_flink_operator_namespace(
        self,
        cli_runner,
        role_arn,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        create_pod_identity_association_response,
        flink_operator_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        result = cli_runner.run(
            [
                "emr-containers",
                "create-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
                "--type",
                "flink_operator",
                "--operator-namespace",
                "flink-operator",
            ]
        )

        assert result.aws_requests[0].service_name == "eks"
        assert result.aws_requests[0].operation_name == "DescribeCluster"
        assert (
            len(result.aws_requests)
            == len(flink_operator_service_accounts) + 1
        )
        for i in range(len(flink_operator_service_accounts)):
            ns = "flink-operator" if i == 0 else namespace
            self.assert_call_matches(
                result.aws_requests[i + 1],
                flink_operator_service_accounts[i],
                cluster_name,
                ns,
                role_arn,
            )

    # Use case: Expect to return create pod identity association results for livy
    # Expected results: Operation is performed by client
    # to create pod identity associations with livy service accounts
    def test_create_role_associations_for_livy(
        self,
        cli_runner,
        role_arn,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        create_pod_identity_association_response,
        livy_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        result = cli_runner.run(
            [
                "emr-containers",
                "create-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
                "--type",
                "livy",
            ]
        )

        assert result.aws_requests[0].service_name == "eks"
        assert result.aws_requests[0].operation_name == "DescribeCluster"
        assert len(result.aws_requests) == len(livy_service_accounts) + 1
        for i in range(len(livy_service_accounts)):
            self.assert_call_matches(
                result.aws_requests[i + 1],
                livy_service_accounts[i],
                cluster_name,
                namespace,
                role_arn,
            )

    # Use case: Expect to return create pod identity association results for livy with controller namespace
    # Expected results: Operation is performed by client
    # to create pod identity associations with livy service accounts in correct namespace
    def test_create_role_associations_for_livy_namespace(
        self,
        cli_runner,
        role_arn,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        create_pod_identity_association_response,
        livy_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        result = cli_runner.run(
            [
                "emr-containers",
                "create-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
                "--type",
                "livy",
                "--operator-namespace",
                "livy",
            ]
        )

        assert result.aws_requests[0].service_name == "eks"
        assert result.aws_requests[0].operation_name == "DescribeCluster"
        assert len(result.aws_requests) == len(livy_service_accounts) + 1
        for i in range(len(livy_service_accounts)):
            ns = "livy" if i == 0 else namespace
            self.assert_call_matches(
                result.aws_requests[i + 1],
                livy_service_accounts[i],
                cluster_name,
                ns,
                role_arn,
            )

    # Use case: Expect to return create pod identity association results for customer input service accounts
    # Expected results: Operation is performed by client
    # to create pod identity associations with customer input service accounts
    def test_create_role_associations_for_customer_service_account(
        self,
        cli_runner,
        role_arn,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        create_pod_identity_association_response,
        spark_operator_service_accounts,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        result = cli_runner.run(
            [
                "emr-containers",
                "create-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
                "--service-account",
                "test_sa",
            ]
        )

        assert result.aws_requests[0].service_name == "eks"
        assert result.aws_requests[0].operation_name == "DescribeCluster"
        assert len(result.aws_requests) == 2
        self.assert_call_matches(
            result.aws_requests[1],
            "test_sa",
            cluster_name,
            namespace,
            role_arn,
        )

    # Use case: Expect to return ResourceInUse exception and do nothing on already existed associations
    # Expected results: Association creations are skipped
    def test_create_role_associations_already_exists_for_start_job_run(
        self,
        cli_runner,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        create_pod_identity_association_already_exists_error_response,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        cli_runner.add_response(
            HTTPResponse(
                body=create_pod_identity_association_already_exists_error_response,
                status_code=409,
            )
        )
        cli_runner.add_response(
            HTTPResponse(
                body=create_pod_identity_association_already_exists_error_response,
                status_code=409,
            )
        )
        cli_runner.add_response(
            HTTPResponse(
                body=create_pod_identity_association_already_exists_error_response,
                status_code=409,
            )
        )
        result = cli_runner.run(
            [
                "emr-containers",
                "create-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
            ]
        )
        components = ["client", "driver", "executor"]
        expected_out = ""
        for component in components:
            expected_out += (
                "Skipping pod identity association creation because pod identity association already exists for service "
                + f"account emr-containers-sa-spark-{component}-123456789012-16o0gwny3p and role myrole in namespace test: "
                + "An error occurred (ResourceInUseException) when calling the "
                + "CreatePodIdentityAssociation operation: Association already exists: a-1\n"
            )
        assert result.stderr == expected_out

    # Use case: Expect to return error exception and rollback on created associations in the same call
    # Expected results: Associations are rolled back
    def test_create_role_associations_error_rollback_for_start_job_run(
        self,
        cli_runner,
        cluster_name,
        namespace,
        role_name,
        describe_cluster_response,
        start_job_run_service_accounts,
        role_arn,
        create_pod_identity_association_response,
        create_pod_identity_association_other_error_response,
        delete_pod_identity_association_response,
    ):
        cli_runner.add_response(HTTPResponse(body=describe_cluster_response))
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        cli_runner.add_response(
            HTTPResponse(body=create_pod_identity_association_response)
        )
        cli_runner.add_response(
            HTTPResponse(
                body=create_pod_identity_association_other_error_response,
                status_code=400,
            )
        )
        for i in range(2):
            cli_runner.add_response(
                HTTPResponse(body=delete_pod_identity_association_response[i])
            )
        result = cli_runner.run(
            [
                "emr-containers",
                "create-role-associations",
                "--cluster-name",
                cluster_name,
                "--namespace",
                namespace,
                "--role-name",
                role_name,
            ]
        )

        assert result.aws_requests[0].service_name == "eks"
        assert result.aws_requests[0].operation_name == "DescribeCluster"
        for i in range(3):
            self.assert_call_matches(
                result.aws_requests[i + 1],
                start_job_run_service_accounts[i],
                cluster_name,
                namespace,
                role_arn,
            )
        for i in range(2):
            self.assert_roll_back(
                result.aws_requests[i + 4], cluster_name, "a-12345678"
            )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from unittest.mock import ANY, Mock, patch

import pytest

from awscli.customizations.ecs.monitormutatinggatewayservice import (
    MUTATION_HANDLERS,
    MonitoringResourcesArgument,
    MonitorMutatingGatewayService,
    MonitorResourcesAction,
    register_monitor_mutating_gateway_service,
)


class TestMonitoringResourcesArgument:
    """Test the custom CLI argument for enabling resource monitoring."""

    def test_add_to_parser(self):
        """Test that the monitoring argument is correctly added to the parser."""
        parser = Mock()
        arg = MonitoringResourcesArgument('monitor-resources')
        arg.add_to_parser(parser)

        parser.add_argument.assert_called_once()
        # Verify the argument was added with correct name
        call_args = parser.add_argument.call_args
        assert call_args[0][0] == '--monitor-resources'
        assert call_args[1]['dest'] == 'monitor_resources'
        assert call_args[1]['choices'] == ['DEPLOYMENT', 'RESOURCE']
        assert call_args[1]['nargs'] == '?'
        # Custom action instead of const
        assert hasattr(call_args[1]['action'], '__call__')


class TestMonitorResourcesAction:
    """Test the custom action for monitor-resources argument."""

    def test_call_with_no_value(self):
        """Test action when flag is provided without value."""
        action = MonitorResourcesAction(None, 'monitor_resources')
        parser = Mock()
        namespace = Mock()

        action(parser, namespace, None, '--monitor-resources')

        assert namespace.monitor_resources == '__DEFAULT__'

    def test_call_with_explicit_value(self):
        """Test action when flag is provided with explicit value."""
        action = MonitorResourcesAction(None, 'monitor_resources')
        parser = Mock()
        namespace = Mock()

        action(parser, namespace, 'RESOURCE', '--monitor-resources')

        assert namespace.monitor_resources == 'RESOURCE'


@pytest.fixture
def mock_watcher_class():
    """Fixture that provides a mock watcher class."""
    watcher_class = Mock()
    watcher_class.is_monitoring_available.return_value = True
    return watcher_class


@pytest.fixture
def handler(mock_watcher_class):
    """Fixture that provides a MonitorMutatingGatewayService handler."""
    return MonitorMutatingGatewayService(
        'create-gateway-service',
        'DEPLOYMENT',
        watcher_class=mock_watcher_class,
    )


class TestMonitorMutatingGatewayService:
    """Test the event handler for monitoring gateway service mutations."""

    def test_init(self, handler):
        """Test proper initialization of the handler."""
        assert handler.api == 'create-gateway-service'
        assert handler.default_resource_view == 'DEPLOYMENT'
        assert handler.api_pascal_case == 'CreateGatewayService'
        assert handler.session is None
        assert handler.parsed_globals is None

    def test_pascal_case_conversion(self):
        """Test API name conversion to PascalCase."""
        test_cases = [
            ('create-gateway-service', 'CreateGatewayService'),
            ('update-gateway-service', 'UpdateGatewayService'),
            ('delete-gateway-service', 'DeleteGatewayService'),
            ('simple-api', 'SimpleApi'),
            ('multi-word-api-name', 'MultiWordApiName'),
        ]

        for api_name, expected_pascal in test_cases:
            handler = MonitorMutatingGatewayService(api_name, 'RESOURCE')
            assert handler.api_pascal_case == expected_pascal

    def test_before_building_argument_table_parser(self, handler):
        """Test storing session for later use."""
        session = Mock()

        handler.before_building_argument_table_parser(session)

        assert handler.session == session

    def test_building_argument_table(self, handler):
        """Test adding monitoring argument to the command's argument table."""
        argument_table = {}
        session = Mock()

        handler.building_argument_table(argument_table, session)

        assert 'monitor-resources' in argument_table
        assert isinstance(
            argument_table['monitor-resources'], MonitoringResourcesArgument
        )

    def test_operation_args_parsed_with_flag(self, handler):
        """Test storing monitoring flag when enabled with default."""
        parsed_args = Mock()
        parsed_args.monitor_resources = '__DEFAULT__'
        parsed_globals = Mock()

        handler.operation_args_parsed(parsed_args, parsed_globals)

        assert handler.effective_resource_view == 'DEPLOYMENT'

    def test_operation_args_parsed_with_explicit_choice(self, handler):
        """Test storing monitoring flag with explicit choice."""
        parsed_args = Mock()
        parsed_args.monitor_resources = 'RESOURCE'
        parsed_globals = Mock()

        handler.operation_args_parsed(parsed_args, parsed_globals)

        assert handler.effective_resource_view == 'RESOURCE'

    def test_operation_args_parsed_without_flag(self, handler):
        """Test storing monitoring flag when disabled."""
        parsed_args = Mock()
        parsed_args.monitor_resources = None
        parsed_args.monitor_mode = None
        parsed_globals = Mock()

        handler.operation_args_parsed(parsed_args, parsed_globals)

        assert handler.effective_resource_view is None

    def test_operation_args_parsed_missing_attribute(self, handler):
        """Test handling missing monitor_resources attribute."""
        # Mock without monitor_resources attribute
        parsed_args = Mock(spec=[])
        parsed_globals = Mock()

        handler.operation_args_parsed(parsed_args, parsed_globals)

        assert handler.effective_resource_view is None

    def test_after_call_monitoring_disabled(self, handler):
        """Test that monitoring is skipped when flag is disabled."""
        handler.effective_resource_view = None
        parsed = {}
        context = Mock()
        http_response = Mock()
        http_response.status_code = 200

        # Should return early without doing anything
        handler.after_call(parsed, context, http_response)

        # No assertions needed - just verify no exceptions

    def test_after_call_http_error(self, handler):
        """Test that monitoring is skipped on HTTP errors."""
        handler.effective_resource_view = 'DEPLOYMENT'
        parsed = {
            'service': {'serviceArn': 'arn:aws:ecs:us-west-2:123:service/test'}
        }
        context = Mock()
        http_response = Mock()
        http_response.status_code = 400

        # Should return early without doing anything
        handler.after_call(parsed, context, http_response)

        # No assertions needed - just verify no exceptions

    def test_after_call_missing_service_arn(self, handler):
        """Test that monitoring is skipped when service ARN is missing."""
        handler.effective_resource_view = 'DEPLOYMENT'
        # Missing serviceArn
        parsed = {'service': {}}
        context = Mock()
        http_response = Mock()
        http_response.status_code = 200

        # Should return early without doing anything
        handler.after_call(parsed, context, http_response)

        # No assertions needed - just verify no exceptions

    def test_after_call_missing_session(self, handler, capsys):
        """Test handling when session is not available."""
        handler.effective_resource_view = 'DEPLOYMENT'
        handler.session = None
        handler.parsed_globals = None

        parsed = {
            'service': {'serviceArn': 'arn:aws:ecs:us-west-2:123:service/test'}
        }
        context = Mock()
        http_response = Mock()
        http_response.status_code = 200

        handler.after_call(parsed, context, http_response)

        captured = capsys.readouterr()
        assert (
            "Unable to create ECS client. Skipping monitoring." in captured.err
        )

    def test_after_call_successful_monitoring(self):
        """Test successful monitoring initiation."""
        # Setup handler state
        mock_watcher_class = Mock()
        mock_watcher = Mock()
        mock_watcher_class.return_value = mock_watcher

        handler = MonitorMutatingGatewayService(
            'create-gateway-service',
            'DEPLOYMENT',
            watcher_class=mock_watcher_class,
        )
        handler.monitor_resources = '__DEFAULT__'
        handler.effective_resource_view = 'DEPLOYMENT'
        handler.effective_mode = 'TEXT-ONLY'

        mock_session = Mock()
        mock_parsed_globals = Mock()
        mock_parsed_globals.region = 'us-west-2'
        mock_parsed_globals.endpoint_url = (
            'https://ecs.us-west-2.amazonaws.com'
        )
        mock_parsed_globals.verify_ssl = True
        mock_parsed_globals.color = 'off'
        handler.session = mock_session
        handler.parsed_globals = mock_parsed_globals

        # Setup mocks
        mock_client = Mock()
        mock_session.create_client.return_value = mock_client

        # Setup call parameters
        service_arn = 'arn:aws:ecs:us-west-2:123456789012:service/test-service'
        parsed = {'service': {'serviceArn': service_arn}}
        context = Mock()
        http_response = Mock()
        http_response.status_code = 200

        # Execute
        handler.after_call(parsed, context, http_response)

        # Verify client creation
        mock_session.create_client.assert_called_once_with(
            'ecs',
            region_name='us-west-2',
            endpoint_url='https://ecs.us-west-2.amazonaws.com',
            verify=True,
        )

        # Verify watcher was created and executed
        mock_watcher_class.assert_called_once_with(
            mock_client,
            service_arn,
            'DEPLOYMENT',
            'TEXT-ONLY',
            use_color=False,
        )
        mock_watcher.exec.assert_called_once()
        # Verify parsed response was cleared
        assert parsed == {}

    @patch('sys.stdout.isatty', return_value=False)
    def test_after_call_interactive_mode_without_tty(
        self, mock_isatty, capsys
    ):
        """Test that interactive mode is skipped when TTY is not available."""
        # Setup handler state
        mock_watcher_class = Mock()

        handler = MonitorMutatingGatewayService(
            'create-gateway-service',
            'DEPLOYMENT',
            watcher_class=mock_watcher_class,
        )
        handler.effective_resource_view = 'DEPLOYMENT'
        handler.effective_mode = 'INTERACTIVE'

        mock_session = Mock()
        mock_parsed_globals = Mock()
        mock_parsed_globals.region = 'us-west-2'
        mock_parsed_globals.endpoint_url = (
            'https://ecs.us-west-2.amazonaws.com'
        )
        mock_parsed_globals.verify_ssl = True
        mock_parsed_globals.color = 'off'
        handler.session = mock_session
        handler.parsed_globals = mock_parsed_globals

        # Setup mocks
        mock_client = Mock()
        mock_session.create_client.return_value = mock_client

        # Setup call parameters
        service_arn = 'arn:aws:ecs:us-west-2:123456789012:service/test-service'
        parsed = {'service': {'serviceArn': service_arn}}
        context = Mock()
        http_response = Mock()
        http_response.status_code = 200

        # Execute
        handler.after_call(parsed, context, http_response)

        # Verify error message was printed
        captured = capsys.readouterr()
        assert (
            "aws: [ERROR]: Interactive mode requires a TTY (terminal)."
            in captured.err
        )
        assert "Use --monitor-mode TEXT-ONLY" in captured.err

        # Verify watcher was not called
        mock_watcher_class.assert_not_called()

    @patch('sys.stdout.isatty', return_value=True)
    def test_after_call_exception_handling(self, mock_isatty, capsys):
        """Test exception handling in after_call method."""
        # Setup handler state
        mock_watcher_class = Mock()
        mock_watcher = Mock()
        mock_watcher.exec.side_effect = Exception("Test exception")
        mock_watcher_class.return_value = mock_watcher

        handler = MonitorMutatingGatewayService(
            'create-gateway-service',
            'DEPLOYMENT',
            watcher_class=mock_watcher_class,
        )
        handler.effective_resource_view = 'DEPLOYMENT'
        handler.effective_mode = 'INTERACTIVE'

        mock_session = Mock()
        mock_parsed_globals = Mock()
        mock_parsed_globals.region = 'us-west-2'
        mock_parsed_globals.endpoint_url = (
            'https://ecs.us-west-2.amazonaws.com'
        )
        mock_parsed_globals.verify_ssl = True
        mock_parsed_globals.color = 'off'
        handler.session = mock_session
        handler.parsed_globals = mock_parsed_globals

        # Setup mocks
        mock_client = Mock()
        mock_session.create_client.return_value = mock_client

        # Setup call parameters
        service_arn = 'arn:aws:ecs:us-west-2:123456789012:service/test-service'
        parsed = {'service': {'serviceArn': service_arn}}
        context = Mock()
        http_response = Mock()
        http_response.status_code = 200

        # Execute - exception should propagate
        with pytest.raises(Exception, match="Test exception"):
            handler.after_call(parsed, context, http_response)

    def test_events(self, handler):
        """Test that correct events are returned for CLI integration."""
        events = handler.events()

        expected_events = [
            "before-building-argument-table-parser.ecs.create-gateway-service",
            "building-argument-table.ecs.create-gateway-service",
            "operation-args-parsed.ecs.create-gateway-service",
            "after-call.ecs.CreateGatewayService",
        ]

        assert len(events) == 4

        for i, (event_name, handler_method) in enumerate(events):
            assert event_name == expected_events[i]
            assert callable(handler_method)


class TestMutationHandlers:
    """Test the predefined mutation handlers."""

    def test_mutation_handlers_configuration(self):
        """Test that mutation handlers are properly configured."""
        assert len(MUTATION_HANDLERS) == 3

        # Test create handler
        create_handler = MUTATION_HANDLERS[0]
        assert create_handler.api == 'create-express-gateway-service'
        assert create_handler.default_resource_view == 'DEPLOYMENT'

        # Test update handler
        update_handler = MUTATION_HANDLERS[1]
        assert update_handler.api == 'update-express-gateway-service'
        assert update_handler.default_resource_view == 'DEPLOYMENT'

        # Test delete handler
        delete_handler = MUTATION_HANDLERS[2]
        assert delete_handler.api == 'delete-express-gateway-service'
        assert delete_handler.default_resource_view == 'RESOURCE'

    def test_multiple_handlers_different_modes(self):
        """Test that different handlers use appropriate monitoring modes."""
        create_handler = MonitorMutatingGatewayService(
            'create-express-gateway-service', 'DEPLOYMENT'
        )
        update_handler = MonitorMutatingGatewayService(
            'update-express-gateway-service', 'DEPLOYMENT'
        )
        delete_handler = MonitorMutatingGatewayService(
            'delete-express-gateway-service', 'RESOURCE'
        )

        # Create and update should use DEPLOYMENT mode for showing new changes
        assert create_handler.default_resource_view == 'DEPLOYMENT'
        assert update_handler.default_resource_view == 'DEPLOYMENT'

        # Delete should use RESOURCE mode to show all resources being removed
        assert delete_handler.default_resource_view == 'RESOURCE'

    def test_api_pascal_case_edge_cases(self):
        """Test PascalCase conversion with edge cases."""
        test_cases = [
            ('a', 'A'),
            ('a-b', 'AB'),
            ('create-express-gateway-service', 'CreateExpressGatewayService'),
            ('api-with-many-dashes', 'ApiWithManyDashes'),
            ('single', 'Single'),
        ]

        for api_name, expected_pascal in test_cases:
            handler = MonitorMutatingGatewayService(api_name, 'RESOURCE')
            assert (
                handler.api_pascal_case == expected_pascal
            ), f"Failed for {api_name} -> {expected_pascal}"


class TestMonitoringIntegration:
    """Test integration between monitoring components."""

    def test_monitoring_resources_argument_integration(self):
        """Test that MonitoringResourcesArgument integrates properly with handlers."""
        handler = MonitorMutatingGatewayService('test-api', 'RESOURCE')

        # Test argument table building
        argument_table = {}
        handler.building_argument_table(argument_table, Mock())

        # Verify the argument was added
        assert 'monitor-resources' in argument_table
        arg = argument_table['monitor-resources']
        assert isinstance(arg, MonitoringResourcesArgument)

        # Test that the argument can be added to a parser
        mock_parser = Mock()
        arg.add_to_parser(mock_parser)

        mock_parser.add_argument.assert_called_once()
        # Verify the argument was added correctly
        call_args = mock_parser.add_argument.call_args
        assert call_args[0][0] == '--monitor-resources'
        assert call_args[1]['dest'] == 'monitor_resources'
        assert call_args[1]['action'] == MonitorResourcesAction


class TestRegisterFunction:
    """Test the registration function for event handlers."""

    def test_register_monitor_mutating_gateway_service(self):
        """Test that all handlers are properly registered."""
        mock_event_handler = Mock()

        register_monitor_mutating_gateway_service(mock_event_handler)

        # Should register 4 events per handler * 3 handlers = 12 total registrations
        assert mock_event_handler.register.call_count == 12

        # Verify some specific event registrations
        registered_events = [
            call[0][0] for call in mock_event_handler.register.call_args_list
        ]

        # Check that create-express-gateway-service events are registered
        assert (
            'before-building-argument-table-parser.ecs.create-express-gateway-service'
            in registered_events
        )
        assert (
            'building-argument-table.ecs.create-express-gateway-service'
            in registered_events
        )
        assert (
            'operation-args-parsed.ecs.create-express-gateway-service'
            in registered_events
        )
        assert (
            'after-call.ecs.CreateExpressGatewayService' in registered_events
        )

        # Check that update-express-gateway-service events are registered
        assert (
            'after-call.ecs.UpdateExpressGatewayService' in registered_events
        )

        # Check that delete-express-gateway-service events are registered
        assert (
            'after-call.ecs.DeleteExpressGatewayService' in registered_events
        )


class TestColorSupport:
    """Test color support functionality in monitormutatinggatewayservice"""

    def test_should_use_color_on(self):
        """Test _should_use_color returns True when color is 'on'"""
        handler = MonitorMutatingGatewayService(
            'create-express-gateway-service', 'DEPLOYMENT'
        )
        parsed_globals = Mock()
        parsed_globals.color = 'on'

        assert handler._should_use_color(parsed_globals) is True

    def test_should_use_color_off(self):
        """Test _should_use_color returns False when color is 'off'"""
        handler = MonitorMutatingGatewayService(
            'create-express-gateway-service', 'DEPLOYMENT'
        )
        parsed_globals = Mock()
        parsed_globals.color = 'off'

        assert handler._should_use_color(parsed_globals) is False

    @pytest.mark.parametrize(
        "isatty_return,expected", [(True, True), (False, False)]
    )
    def test_should_use_color_auto(self, isatty_return, expected):
        """Test _should_use_color with 'auto' setting"""
        with patch('sys.stdout.isatty', return_value=isatty_return):
            handler = MonitorMutatingGatewayService(
                'create-express-gateway-service', 'DEPLOYMENT'
            )
            parsed_globals = Mock()
            parsed_globals.color = 'auto'

            assert handler._should_use_color(parsed_globals) is expected



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.compat import StringIO
from awscli.customizations.configservice.getstatus import GetStatusCommand
from awscli.testutils import mock, unittest


class TestGetStatusCommand(unittest.TestCase):
    def setUp(self):
        self.session = mock.Mock()

        # Setup the config client mock.
        self.config_client = mock.Mock()
        self.session.create_client.return_value = self.config_client

        # Create some handles to control the description outputs.
        self.recorder_status = []
        self.channel_status = []

        # Set the output handles to the client.
        self.config_client.describe_configuration_recorder_status.return_value = {
            'ConfigurationRecordersStatus': self.recorder_status
        }
        self.config_client.describe_delivery_channel_status.return_value = {
            'DeliveryChannelsStatus': self.channel_status
        }

        self.parsed_args = mock.Mock()
        self.parsed_globals = mock.Mock()
        self.cmd = GetStatusCommand(self.session)

    def _make_delivery_channel_status(
        self,
        name,
        stream_delivery_status,
        history_delivery_status,
        snapshot_delivery_status,
    ):
        status = {
            'name': 'default',
            'configStreamDeliveryInfo': stream_delivery_status,
            'configHistoryDeliveryInfo': history_delivery_status,
            'configSnapshotDeliveryInfo': snapshot_delivery_status,
        }
        return status

    def test_create_client(self):
        # Set values for the parsed globals.
        self.parsed_globals.verify_ssl = True
        self.parsed_globals.region = 'us-east-1'
        self.parsed_globals.endpoint_url = 'https://foo.com'

        # Ensure that the client was built with the proper arguments.
        self.cmd._run_main(self.parsed_args, self.parsed_globals)
        self.session.create_client.assert_called_with(
            'config',
            verify=self.parsed_globals.verify_ssl,
            region_name=self.parsed_globals.region,
            endpoint_url=self.parsed_globals.endpoint_url,
        )

    def test_configuration_recorder_success(self):
        status = {
            'name': 'default',
            'recording': True,
            'lastStatus': 'SUCCESS',
        }
        self.recorder_status.append(status)

        expected_output = (
            'Configuration Recorders:\n\n'
            'name: default\n'
            'recorder: ON\n'
            'last status: SUCCESS\n\n'
            'Delivery Channels:\n\n'
        )

        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            self.cmd._run_main(self.parsed_args, self.parsed_globals)
            self.assertEqual(expected_output, mock_stdout.getvalue())

    def test_configuration_recorder_fail(self):
        status = {
            'name': 'default',
            'recording': True,
            'lastStatus': 'FAILURE',
            'lastErrorCode': '500',
            'lastErrorMessage': 'This is the error',
        }
        self.recorder_status.append(status)

        expected_output = (
            'Configuration Recorders:\n\n'
            'name: default\n'
            'recorder: ON\n'
            'last status: FAILURE\n'
            'error code: 500\n'
            'message: This is the error\n\n'
            'Delivery Channels:\n\n'
        )

        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            self.cmd._run_main(self.parsed_args, self.parsed_globals)
            self.assertEqual(expected_output, mock_stdout.getvalue())

    def test_configuration_recorder_off(self):
        status = {'name': 'default', 'recording': False}
        self.recorder_status.append(status)

        expected_output = (
            'Configuration Recorders:\n\n'
            'name: default\n'
            'recorder: OFF\n\n'
            'Delivery Channels:\n\n'
        )

        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            self.cmd._run_main(self.parsed_args, self.parsed_globals)
            self.assertEqual(expected_output, mock_stdout.getvalue())

    def test_multiple_configuration_recorders(self):
        status = {
            'name': 'default',
            'recording': True,
            'lastStatus': 'SUCCESS',
        }
        self.recorder_status.append(status)

        status = {
            'name': 'default',
            'recording': True,
            'lastStatus': 'FAILURE',
            'lastErrorCode': '500',
            'lastErrorMessage': 'This is the error',
        }
        self.recorder_status.append(status)

        status = {'name': 'default', 'recording': False}
        self.recorder_status.append(status)

        expected_output = (
            'Configuration Recorders:\n\n'
            'name: default\n'
            'recorder: ON\n'
            'last status: SUCCESS\n\n'
            'name: default\n'
            'recorder: ON\n'
            'last status: FAILURE\n'
            'error code: 500\n'
            'message: This is the error\n\n'
            'name: default\n'
            'recorder: OFF\n\n'
            'Delivery Channels:\n\n'
        )
        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            self.cmd._run_main(self.parsed_args, self.parsed_globals)
            self.assertEqual(expected_output, mock_stdout.getvalue())

    def test_delivery_channel_success_single_delivery_info(self):
        name = 'default'
        success = {'lastStatus': 'SUCCESS'}

        stream_delivery_status = success
        history_delivery_status = {}
        snapshot_delivery_status = {}

        status = self._make_delivery_channel_status(
            name,
            stream_delivery_status=stream_delivery_status,
            history_delivery_status=history_delivery_status,
            snapshot_delivery_status=snapshot_delivery_status,
        )
        self.channel_status.append(status)

        expected_output = (
            'Configuration Recorders:\n\n'
            'Delivery Channels:\n\n'
            'name: default\n'
            'last stream delivery status: SUCCESS\n\n'
        )

        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            self.cmd._run_main(self.parsed_args, self.parsed_globals)
            self.assertEqual(expected_output, mock_stdout.getvalue())

    def test_delivery_channel_success_multiple_delivery_info(self):
        name = 'default'
        success = {'lastStatus': 'SUCCESS'}

        stream_delivery_status = success
        history_delivery_status = success
        snapshot_delivery_status = success

        status = self._make_delivery_channel_status(
            name,
            stream_delivery_status=stream_delivery_status,
            history_delivery_status=history_delivery_status,
            snapshot_delivery_status=snapshot_delivery_status,
        )
        self.channel_status.append(status)

        expected_output = (
            'Configuration Recorders:\n\n'
            'Delivery Channels:\n\n'
            'name: default\n'
            'last stream delivery status: SUCCESS\n'
            'last history delivery status: SUCCESS\n'
            'last snapshot delivery status: SUCCESS\n\n'
        )

        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            self.cmd._run_main(self.parsed_args, self.parsed_globals)
            self.assertEqual(expected_output, mock_stdout.getvalue())

    def test_delivery_channel_fail_single_delivery_info(self):
        name = 'default'
        failure = {
            'lastStatus': 'FAILURE',
            'lastErrorCode': '500',
            'lastErrorMessage': 'This is the error',
        }

        stream_delivery_status = failure
        history_delivery_status = {}
        snapshot_delivery_status = {}

        status = self._make_delivery_channel_status(
            name,
            stream_delivery_status=stream_delivery_status,
            history_delivery_status=history_delivery_status,
            snapshot_delivery_status=snapshot_delivery_status,
        )
        self.channel_status.append(status)

        expected_output = (
            'Configuration Recorders:\n\n'
            'Delivery Channels:\n\n'
            'name: default\n'
            'last stream delivery status: FAILURE\n'
            'error code: 500\n'
            'message: This is the error\n\n'
        )

        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            self.cmd._run_main(self.parsed_args, self.parsed_globals)
            self.assertEqual(expected_output, mock_stdout.getvalue())

    def test_delivery_channel_mixed_multiple_delivery_info(self):
        name = 'default'
        success = {'lastStatus': 'SUCCESS'}
        failure = {
            'lastStatus': 'FAILURE',
            'lastErrorCode': '500',
            'lastErrorMessage': 'This is the error',
        }

        stream_delivery_status = failure
        history_delivery_status = success
        snapshot_delivery_status = success

        status = self._make_delivery_channel_status(
            name,
            stream_delivery_status=stream_delivery_status,
            history_delivery_status=history_delivery_status,
            snapshot_delivery_status=snapshot_delivery_status,
        )
        self.channel_status.append(status)

        expected_output = (
            'Configuration Recorders:\n\n'
            'Delivery Channels:\n\n'
            'name: default\n'
            'last stream delivery status: FAILURE\n'
            'error code: 500\n'
            'message: This is the error\n'
            'last history delivery status: SUCCESS\n'
            'last snapshot delivery status: SUCCESS\n\n'
        )

        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            self.cmd._run_main(self.parsed_args, self.parsed_globals)
            self.assertEqual(expected_output, mock_stdout.getvalue())

    def test_multiple_delivery_channels(self):
        name = 'default'
        success = {'lastStatus': 'SUCCESS'}
        failure = {
            'lastStatus': 'FAILURE',
            'lastErrorCode': '500',
            'lastErrorMessage': 'This is the error',
        }

        stream_delivery_status = failure
        history_delivery_status = success
        snapshot_delivery_status = success

        status = self._make_delivery_channel_status(
            name,
            stream_delivery_status=stream_delivery_status,
            history_delivery_status=history_delivery_status,
            snapshot_delivery_status=snapshot_delivery_status,
        )
        self.channel_status.append(status)
        self.channel_status.append(status)

        expected_output = (
            'Configuration Recorders:\n\n'
            'Delivery Channels:\n\n'
            'name: default\n'
            'last stream delivery status: FAILURE\n'
            'error code: 500\n'
            'message: This is the error\n'
            'last history delivery status: SUCCESS\n'
            'last snapshot delivery status: SUCCESS\n\n'
            'name: default\n'
            'last stream delivery status: FAILURE\n'
            'error code: 500\n'
            'message: This is the error\n'
            'last history delivery status: SUCCESS\n'
            'last snapshot delivery status: SUCCESS\n\n'
        )

        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            self.cmd._run_main(self.parsed_args, self.parsed_globals)
            self.assertEqual(expected_output, mock_stdout.getvalue())

    def test_full_get_status(self):
        # Create the configuration recorder statuses.
        status = {
            'name': 'default',
            'recording': True,
            'lastStatus': 'SUCCESS',
        }
        self.recorder_status.append(status)

        status = {
            'name': 'default',
            'recording': True,
            'lastStatus': 'FAILURE',
            'lastErrorCode': '500',
            'lastErrorMessage': 'This is the error',
        }
        self.recorder_status.append(status)

        status = {'name': 'default', 'recording': False}
        self.recorder_status.append(status)

        # Create the delivery channel statuses.
        name = 'default'
        success = {'lastStatus': 'SUCCESS'}
        failure = {
            'lastStatus': 'FAILURE',
            'lastErrorCode': '500',
            'lastErrorMessage': 'This is the error',
        }

        stream_delivery_status = failure
        history_delivery_status = success
        snapshot_delivery_status = success

        status = self._make_delivery_channel_status(
            name,
            stream_delivery_status=stream_delivery_status,
            history_delivery_status=history_delivery_status,
            snapshot_delivery_status=snapshot_delivery_status,
        )
        self.channel_status.append(status)
        self.channel_status.append(status)

        expected_output = (
            'Configuration Recorders:\n\n'
            'name: default\n'
            'recorder: ON\n'
            'last status: SUCCESS\n\n'
            'name: default\n'
            'recorder: ON\n'
            'last status: FAILURE\n'
            'error code: 500\n'
            'message: This is the error\n\n'
            'name: default\n'
            'recorder: OFF\n\n'
            'Delivery Channels:\n\n'
            'name: default\n'
            'last stream delivery status: FAILURE\n'
            'error code: 500\n'
            'message: This is the error\n'
            'last history delivery status: SUCCESS\n'
            'last snapshot delivery status: SUCCESS\n\n'
            'name: default\n'
            'last stream delivery status: FAILURE\n'
            'error code: 500\n'
            'message: This is the error\n'
            'last history delivery status: SUCCESS\n'
            'last snapshot delivery status: SUCCESS\n\n'
        )

        with mock.patch('sys.stdout', StringIO()) as mock_stdout:
            self.cmd._run_main(self.parsed_args, self.parsed_globals)
            self.assertEqual(expected_output, mock_stdout.getvalue())



# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.exceptions import ProfileNotFound
from botocore.session import Session

from awscli.customizations import timestampformat
from awscli.customizations.exceptions import ConfigurationError
from awscli.testutils import mock, unittest


class TestScalarParse(unittest.TestCase):
    def test_register_timestamp_format(self):
        event_handers = mock.Mock()
        timestampformat.register_timestamp_format(event_handers)
        event_handers.register_first.assert_called_with(
            'session-initialized', timestampformat.add_timestamp_parser
        )

    def test_identity(self):
        self.assertEqual(timestampformat.identity('foo'), 'foo')
        self.assertEqual(timestampformat.identity(10), 10)

    def test_scalar_parsers_set(self):
        session = mock.Mock()
        session.get_scoped_config.return_value = {
            'cli_timestamp_format': 'wire'
        }
        timestampformat.add_timestamp_parser(session)
        session.get_component.assert_called_with('response_parser_factory')
        factory = session.get_component.return_value
        expected = [mock.call(timestamp_parser=timestampformat.identity)]
        self.assertEqual(factory.set_parser_defaults.mock_calls, expected)

    def test_choose_none_timestamp_formatter(self):
        session = mock.Mock(spec=Session)
        session.get_scoped_config.return_value = {
            'cli_timestamp_format': 'wire'
        }
        factory = session.get_component.return_value
        timestampformat.add_timestamp_parser(session)
        factory.set_parser_defaults.assert_called_with(
            timestamp_parser=timestampformat.identity
        )

    def test_choose_iso_timestamp_formatter(self):
        session = mock.Mock(spec=Session)
        session.get_scoped_config.return_value = {
            'cli_timestamp_format': 'iso8601'
        }
        factory = session.get_component.return_value
        timestampformat.add_timestamp_parser(session)
        factory.set_parser_defaults.assert_called_with(
            timestamp_parser=timestampformat.iso_format
        )

    def test_choose_invalid_timestamp_formatter(self):
        session = mock.Mock(spec=Session)
        session.get_scoped_config.return_value = {
            'cli_timestamp_format': 'foobar'
        }
        session.get_component.return_value
        with self.assertRaises(ConfigurationError):
            timestampformat.add_timestamp_parser(session)

    def test_choose_timestamp_parser_profile_not_found(self):
        session = mock.Mock(spec=Session)
        session.get_scoped_config.side_effect = ProfileNotFound(profile='foo')
        factory = session.get_component.return_value
        timestampformat.add_timestamp_parser(session)
        factory.set_parser_defaults.assert_called_with(
            timestamp_parser=timestampformat.iso_format
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.arguments import CLIArgument
from awscli.clidriver import create_clidriver
from awscli.customizations.configservice.putconfigurationrecorder import (
    ConfigurationRecorderArgument,
    RecordingGroupArgument,
    extract_recording_group,
)
from awscli.testutils import unittest


class TestPutConfigurationRecorder(unittest.TestCase):
    def setUp(self):
        self.clidriver = create_clidriver()
        self.session = self.clidriver.session
        self.argument_table = {}
        self.service_model = self.session.get_service_model('config')
        self.operation_model = self.service_model.operation_model(
            'PutConfigurationRecorder'
        )
        configuration_recorder_model = (
            self.operation_model.input_shape.members['ConfigurationRecorder']
        )
        self.old_configuration_recorder_argument = CLIArgument(
            name='configuration-recorder',
            argument_model=configuration_recorder_model,
            operation_model=self.operation_model,
            is_required=True,
            event_emitter=self.session.get_component('event_emitter'),
            serialized_name='ConfigurationRecorder',
        )
        self.argument_table['configuration-recorder'] = (
            self.old_configuration_recorder_argument
        )

    def test_extract_recording_group(self):
        extract_recording_group(self.session, self.argument_table)
        self.assertEqual(len(self.argument_table), 2)

        # Ensure the original argument was replaced with the updated argument.
        self.assertIn('configuration-recorder', self.argument_table)
        new_configuration_recorder_argument = self.argument_table[
            'configuration-recorder'
        ]
        self.assertIsNot(
            new_configuration_recorder_argument,
            self.old_configuration_recorder_argument,
        )
        self.assertIsInstance(
            new_configuration_recorder_argument, ConfigurationRecorderArgument
        )

        # Ensure the recording group member was extracted to an argument
        self.assertIn('recording-group', self.argument_table)
        recording_group_argument = self.argument_table['recording-group']
        self.assertIsInstance(recording_group_argument, RecordingGroupArgument)

    def test_configuration_recorder_when_new_value(self):
        value = '{"name":"myname","roleARN":"myarn"}'
        parameters = {}
        extract_recording_group(self.session, self.argument_table)
        configuration_recorder_argument = self.argument_table[
            'configuration-recorder'
        ]
        configuration_recorder_argument.add_to_params(parameters, value)
        self.assertEqual(
            {'ConfigurationRecorder': {'name': 'myname', 'roleARN': 'myarn'}},
            parameters,
        )

    def test_configuration_recorder_when_update_value(self):
        value = '{"name":"myname","roleARN":"myarn"}'
        parameters = {
            'ConfigurationRecorder': {
                'recordingGroup': {
                    'allSupported': True,
                    'resourceTypes': ['AWS::EC2::Volume'],
                }
            }
        }
        extract_recording_group(self.session, self.argument_table)
        configuration_recorder_argument = self.argument_table[
            'configuration-recorder'
        ]
        configuration_recorder_argument.add_to_params(parameters, value)
        self.assertEqual(
            {
                'ConfigurationRecorder': {
                    'name': 'myname',
                    'roleARN': 'myarn',
                    'recordingGroup': {
                        'allSupported': True,
                        'resourceTypes': ['AWS::EC2::Volume'],
                    },
                }
            },
            parameters,
        )

    def test_recording_group_when_new_value(self):
        value = '{"allSupported":true,"resourceTypes":["AWS::EC2::Volume"]}'
        parameters = {}
        extract_recording_group(self.session, self.argument_table)
        recording_group_argument = self.argument_table['recording-group']
        recording_group_argument.add_to_params(parameters, value)
        self.assertEqual(
            {
                'ConfigurationRecorder': {
                    'recordingGroup': {
                        'allSupported': True,
                        'resourceTypes': ['AWS::EC2::Volume'],
                    }
                }
            },
            parameters,
        )

    def test_recording_group_when_update_value(self):
        value = '{"allSupported":true,"resourceTypes":["AWS::EC2::Volume"]}'
        parameters = {
            'ConfigurationRecorder': {
                'name': 'myname',
                'roleARN': 'myarn',
            }
        }
        extract_recording_group(self.session, self.argument_table)
        recording_group_argument = self.argument_table['recording-group']
        recording_group_argument.add_to_params(parameters, value)
        self.assertEqual(
            {
                'ConfigurationRecorder': {
                    'name': 'myname',
                    'roleARN': 'myarn',
                    'recordingGroup': {
                        'allSupported': True,
                        'resourceTypes': ['AWS::EC2::Volume'],
                    },
                }
            },
            parameters,
        )



import os
import sys
from contextlib import contextmanager

import colorama
from colorama import Back, Fore

from awscli.compat import StringIO
from awscli.table import COLORAMA_KWARGS
from awscli.testutils import capture_output, skip_if_windows, unittest


@skip_if_windows('Posix color code tests')
class TestPosix(unittest.TestCase):
    _ANSI_RESET_ALL = '\033[0m'
    _ANSI_FORE_RED = '\033[31m'
    _ANSI_FORE_BLUE = '\033[34m'

    def test_fore_attributes(self):
        self.assertEqual(Fore.BLACK, '\033[30m')
        self.assertEqual(Fore.RED, '\033[31m')
        self.assertEqual(Fore.GREEN, '\033[32m')
        self.assertEqual(Fore.YELLOW, '\033[33m')
        self.assertEqual(Fore.BLUE, '\033[34m')
        self.assertEqual(Fore.MAGENTA, '\033[35m')
        self.assertEqual(Fore.CYAN, '\033[36m')
        self.assertEqual(Fore.WHITE, '\033[37m')
        self.assertEqual(Fore.RESET, '\033[39m')

        # Check the light, extended versions.
        self.assertEqual(Fore.LIGHTBLACK_EX, '\033[90m')
        self.assertEqual(Fore.LIGHTRED_EX, '\033[91m')
        self.assertEqual(Fore.LIGHTGREEN_EX, '\033[92m')
        self.assertEqual(Fore.LIGHTYELLOW_EX, '\033[93m')
        self.assertEqual(Fore.LIGHTBLUE_EX, '\033[94m')
        self.assertEqual(Fore.LIGHTMAGENTA_EX, '\033[95m')
        self.assertEqual(Fore.LIGHTCYAN_EX, '\033[96m')
        self.assertEqual(Fore.LIGHTWHITE_EX, '\033[97m')

    def test_back_attributes(self):
        self.assertEqual(Back.BLACK, '\033[40m')
        self.assertEqual(Back.RED, '\033[41m')
        self.assertEqual(Back.GREEN, '\033[42m')
        self.assertEqual(Back.YELLOW, '\033[43m')
        self.assertEqual(Back.BLUE, '\033[44m')
        self.assertEqual(Back.MAGENTA, '\033[45m')
        self.assertEqual(Back.CYAN, '\033[46m')
        self.assertEqual(Back.WHITE, '\033[47m')
        self.assertEqual(Back.RESET, '\033[49m')

        # Check the light, extended versions.
        self.assertEqual(Back.LIGHTBLACK_EX, '\033[100m')
        self.assertEqual(Back.LIGHTRED_EX, '\033[101m')
        self.assertEqual(Back.LIGHTGREEN_EX, '\033[102m')
        self.assertEqual(Back.LIGHTYELLOW_EX, '\033[103m')
        self.assertEqual(Back.LIGHTBLUE_EX, '\033[104m')
        self.assertEqual(Back.LIGHTMAGENTA_EX, '\033[105m')
        self.assertEqual(Back.LIGHTCYAN_EX, '\033[106m')
        self.assertEqual(Back.LIGHTWHITE_EX, '\033[107m')

    @contextmanager
    def colorama_text(self, tty=True):
        with capture_output() as captured:
            captured.stdout.isatty = lambda: tty
            with colorama.colorama_text(**COLORAMA_KWARGS):
                yield captured

    def test_colorama_auto_resets(self):
        with self.colorama_text() as captured:
            content = Fore.RED + 'foo'
            sys.stdout.write(content)
            sys.stdout.write('bar')
        # Since auto-reset is enabled each call to write should end with the
        # reset code. And foo should start with the red foreground code. This
        # also depends on the strip=False behavior to pass on windows.
        self.assertEqual(
            captured.stdout.getvalue(),
            '%sfoo%sbar%s'
            % (
                self._ANSI_FORE_RED,
                self._ANSI_RESET_ALL,
                self._ANSI_RESET_ALL,
            ),
        )

    def test_colorama_does_not_strip(self):
        # Strip is set to False which means that it doesn't remove ansi codes
        # from the output if its a pipe.
        with self.colorama_text() as captured:
            content = Fore.BLUE + 'foo'
            sys.stdout.write(content)
            self.assertEqual(
                captured.stdout.getvalue(),
                '%sfoo%s' % (self._ANSI_FORE_BLUE, self._ANSI_RESET_ALL),
            )

    def test_colorama_does_not_strip_non_tty(self):
        with self.colorama_text(tty=False) as captured:
            content = Fore.BLUE + 'foo'
            sys.stdout.write(content)
            self.assertEqual(
                captured.stdout.getvalue(),
                '%sfoo%s' % (self._ANSI_FORE_BLUE, self._ANSI_RESET_ALL),
            )
